@@ -1,19 +1,26 @@
 import itertools
+import os, shutil
 import warnings
 from collections.abc import Callable, Sequence
 from itertools import product
 from typing import Union
+from tqdm import tqdm
 
 import dask.array as da
 import numpy as np
+import zarr
 import spatial_image as si
 import xarray as xr
 from dask import delayed
 from dask.utils import has_keyword
+from dask.array.core import normalize_chunks
+from dask import config as dask_config
 
 from multiview_stitcher import (
     mv_graph,
+    ngff_utils,
     param_utils,
+    misc_utils,
     transformation,
     weights,
 )
@@ -114,6 +121,60 @@ def simple_average_fusion(
     ).astype(transformed_views[0].dtype)
 
 
+def process_output_chunksize(sims, output_chunksize):
+
+    ndim = si_utils.get_ndim_from_sim(sims[0])
+    sdims = si_utils.get_spatial_dims_from_sim(sims[0])
+
+    if output_chunksize is None:
+        if isinstance(sims[0].data, da.Array):
+            # if first tile is a chunked dask array, use its chunksize
+            output_chunksize = dict(zip(sdims, sims[0].data.chunksize[-ndim:]))
+        else:
+            # if first tile is not a chunked dask array, use default chunksize
+            # defined in spatial_image_utils.py
+            output_chunksize = si_utils.get_default_spatial_chunksizes(ndim)
+    elif isinstance(output_chunksize, int):
+        output_chunksize = {dim: output_chunksize for dim in sdims}
+
+    return output_chunksize
+
+
+def process_output_stack_properties(
+    sims,
+    output_spacing,
+    output_origin,
+    output_shape,
+    output_stack_properties,
+    output_stack_mode,
+    transform_key,
+):
+    
+    params = [
+        si_utils.get_affine_from_sim(sim, transform_key=transform_key)
+        for sim in sims
+    ]
+
+    if output_stack_properties is None:
+        if output_spacing is None:
+            output_spacing = si_utils.get_spacing_from_sim(sims[0])
+
+        output_stack_properties = calc_fusion_stack_properties(
+            sims,
+            params=params,
+            spacing=output_spacing,
+            mode=output_stack_mode,
+        )
+
+        if output_origin is not None:
+            output_stack_properties["origin"] = output_origin
+
+        if output_shape is not None:
+            output_stack_properties["shape"] = output_shape
+
+    return output_stack_properties
+
+
 def fuse(
     sims: list,
     transform_key: str = None,
@@ -130,6 +191,9 @@ def fuse(
     overlap_in_pixels: int = None,
     interpolation_order: int = 1,
     blending_widths: dict[str, float] = None,
+    output_zarr_url: str | None = None,
+    zarr_options: dict | None = None,
+    batch_options: dict | None = None,
 ):
     """
 
@@ -173,50 +237,149 @@ def fuse(
         Chunksize of the dask data array of the fused image. If the first tile is a chunked dask array,
         its chunksize is used as the default. If the first tile is not a chunked dask array,
         the default chunksize defined in spatial_image_utils.py is used.
-
+    output_zarr_url : str or None, optional
+        If not None, fuse directly into a Zarr store at this location and do so in batches of chunks,
+        with each chunk being processed independently. This allows for efficient memory usage and
+        works well for very large datasets (successfully tested ~0.5PB on a macbook).
+        When provided, fuse() performs eager fusion and returns a SpatialImage backed by the written store.
+    zarr_options: dict, optional
+        Additional (dict of) options to pass when creating the Zarr store. Keys:
+        - ome_zarr : bool, optional
+            If True and output_zarr_url is provided, write a NGFF/OME-Zarr multiscale image under
+            "<output_zarr_url>/". Otherwise, the fused array is written directly under output_zarr_url.
+        - ngff_version : str, optional
+            NGFF version used when ome_zarr=True. Default "0.4".
+        - zarr_array_creation_kwargs: dict = None, optional
+            Additional keyword arguments to pass when creating the Zarr array.
+        - overwrite: bool, by default True
+    batch_options : dict, optional
+        Options for chunked fusion when output_zarr_url is provided. Keys:
+        - batch_func: Callable, optional
+            Function to process each batch of fused chunks. Inputs:
+            1) a list of block_id(s)
+            2) function that performs fusion when passed a given block_id
+            By default None, in which case the each block is processed sequentially.
+        - n_batch: int
+            Number of blocks to process in each batch, by default 1
+        - batch_func_kwargs: dict, optional
+            Additional keyword arguments passed to batch_func.
     Returns
     -------
     SpatialImage
         Fused image.
     """
+    # If writing directly to Zarr/OME-Zarr, run chunked fusion path and return eagerly.
+    if output_zarr_url is not None:
+        # Collect batch options with defaults
+        batch_options = batch_options or {}
+        batch_func = batch_options.get("batch_func", None)
+        n_batch = batch_options.get("n_batch", 1)
+        batch_func_kwargs = batch_options.get("batch_func_kwargs", None)
+        zarr_array_creation_kwargs = batch_options.get("zarr_array_creation_kwargs", None)
+
+        # Collect zarr options with defaults
+        zarr_options = zarr_options or {}
+        ome_zarr = zarr_options.get("ome_zarr", False)
+        ngff_version = zarr_options.get("ngff_version", "0.4")
+        overwrite = zarr_options.get("overwrite", True)
+
+        # Resolve store path for data (OME-Zarr stores scale 0 under "<root>/0")
+        store_url = os.path.join(output_zarr_url, "0") if ome_zarr else output_zarr_url
+
+        if overwrite and os.path.exists(store_url):
+            shutil.rmtree(store_url)
+        if ome_zarr:
+            # Ensure creation kwargs reflect NGFF version when writing OME-Zarr
+            zarr_array_creation_kwargs = ngff_utils.update_zarr_array_creation_kwargs_for_ngff_version(
+                ngff_version, zarr_array_creation_kwargs
+            )
+
+        # Build kwargs for per-chunk fuse() calls (exclude zarr-specific args to avoid recursion)
+        per_chunk_fuse_kwargs = {
+            "sims": sims,
+            "transform_key": transform_key,
+            "fusion_func": fusion_func,
+            "fusion_method_kwargs": fusion_method_kwargs,
+            "weights_func": weights_func,
+            "weights_func_kwargs": weights_func_kwargs,
+            "output_spacing": output_spacing,
+            "output_stack_mode": output_stack_mode,
+            "output_origin": output_origin,
+            "output_shape": output_shape,
+            "output_stack_properties": output_stack_properties,
+            "output_chunksize": output_chunksize,
+            "overlap_in_pixels": overlap_in_pixels,
+            "interpolation_order": interpolation_order,
+            "blending_widths": blending_widths,
+        }
+
+        # Prepare block fusion and process in batches
+        block_fusion_info = prepare_block_fusion(
+            store_url,
+            fuse_kwargs=per_chunk_fuse_kwargs,
+            zarr_array_creation_kwargs=zarr_array_creation_kwargs,
+        )
+        fuse_chunk = block_fusion_info["func"]
+        nblocks = block_fusion_info["nblocks"]
+        osp = block_fusion_info["output_stack_properties"]
+        osp["shape"] = {dim: int(v) for dim, v in osp["shape"].items()}
+
+        print(f'Fusing {np.prod(nblocks)} blocks in batches of {n_batch}...')
+        for batch in tqdm(
+            misc_utils.ndindex_batches(nblocks, n_batch),
+            total=int(np.ceil(np.prod(nblocks) / n_batch)),
+        ):
+            if batch_func is None:
+                for block_id in batch:
+                    fuse_chunk(block_id)
+            else:
+                batch_func(fuse_chunk, batch, **(batch_func_kwargs or {}))
+
+        # Build SpatialImage from zarr array
+        z = zarr.open(store_url, mode="r")
+        fusion_transform_key = transform_key
+        fused = si_utils.get_sim_from_array(
+            array=z,
+            dims=list(sims[0].dims),
+            transform_key=fusion_transform_key,
+            scale=osp["spacing"],
+            translation=osp["origin"],
+            c_coords=sims[0].coords["c"].values,
+            t_coords=sims[0].coords["t"].values,
+        )
+
+        # If requested, write OME-Zarr metadata
+        # and multiscale pyramid
+        if ome_zarr:
+            ngff_utils.write_sim_to_ome_zarr(
+                fused,
+                output_zarr_url=output_zarr_url,
+                overwrite=False,
+            )
+
+        return fused
+
+    # Default in-memory fusion path (unchanged)
+    output_chunksize = process_output_chunksize(sims, output_chunksize)
+
+    output_stack_properties = process_output_stack_properties(
+        sims=sims,
+        output_spacing=output_spacing,
+        output_origin=output_origin,
+        output_shape=output_shape,
+        output_stack_properties=output_stack_properties,
+        output_stack_mode=output_stack_mode,
+        transform_key=transform_key,
+    )
 
-    ndim = si_utils.get_ndim_from_sim(sims[0])
     sdims = si_utils.get_spatial_dims_from_sim(sims[0])
-    nsdims = [dim for dim in sims[0].dims if dim not in sdims]
-
-    if output_chunksize is None:
-        if isinstance(sims[0].data, da.Array):
-            # if first tile is a chunked dask array, use its chunksize
-            output_chunksize = dict(zip(sdims, sims[0].data.chunksize[-ndim:]))
-        else:
-            # if first tile is not a chunked dask array, use default chunksize
-            # defined in spatial_image_utils.py
-            output_chunksize = si_utils.get_default_spatial_chunksizes(ndim)
-    elif isinstance(output_chunksize, int):
-        output_chunksize = {dim: output_chunksize for dim in sdims}
+    nsdims = si_utils.get_nonspatial_dims_from_sim(sims[0])
 
     params = [
         si_utils.get_affine_from_sim(sim, transform_key=transform_key)
         for sim in sims
     ]
 
-    if output_stack_properties is None:
-        if output_spacing is None:
-            output_spacing = si_utils.get_spacing_from_sim(sims[0])
-
-        output_stack_properties = calc_fusion_stack_properties(
-            sims,
-            params=params,
-            spacing=output_spacing,
-            mode=output_stack_mode,
-        )
-
-        if output_origin is not None:
-            output_stack_properties["origin"] = output_origin
-
-        if output_shape is not None:
-            output_stack_properties["shape"] = output_shape
-
     # determine overlap from weights method
     # (soon: fusion methods will also require overlap)
     overlap_in_pixels = 0
@@ -907,3 +1070,150 @@ def get_interpolated_image(
     interp_image[missing_y, missing_x] = interp_values
 
     return interp_image
+
+
+def prepare_block_fusion(
+    output_zarr_url: str,
+    fuse_kwargs: dict,
+    zarr_array_creation_kwargs: dict = None,
+):
+    """
+    Prepare chunkwise fusion function and number of blocks
+    for embarrassingly parallel fusion
+    """
+
+    sims = fuse_kwargs.get("sims")
+
+    output_stack_properties = process_output_stack_properties(
+        sims=sims,
+        output_stack_properties=fuse_kwargs.pop("output_stack_properties", None),
+        output_spacing=fuse_kwargs.pop("output_spacing", None),
+        output_origin=fuse_kwargs.pop("output_origin", None),
+        output_shape=fuse_kwargs.pop("output_shape", None),
+        output_stack_mode=fuse_kwargs.pop("output_stack_mode", "union"),
+        transform_key=fuse_kwargs.get("transform_key", None)
+    )
+
+    output_chunksize = process_output_chunksize(
+        sims, fuse_kwargs.get("output_chunksize", None)
+    )
+
+    dims = sims[0].dims
+    nsdims = si_utils.get_nonspatial_dims_from_sim(sims[0])
+    sdims = si_utils.get_spatial_dims_from_sim(sims[0])
+    ns_shape = {dim: len(sims[0].coords[dim]) for dim in nsdims}
+
+    full_output_shape = [ns_shape[dim] for dim in nsdims]\
+         + [output_stack_properties['shape'][dim] for dim in sdims]
+    full_output_chunksize = [1,] * len(nsdims)\
+         + [int(output_chunksize[dim]) for dim in sdims]
+    
+    normalized_chunks = normalize_chunks(
+        shape=full_output_shape,
+        chunks=full_output_chunksize)
+    
+    print(f"Fusing into a an output stack:")
+    print("- shape: ", {dim: int(output_stack_properties['shape'][dim])
+        if dim in sdims else 1 for dim in dims})
+    print("- spacing: ", {k: float(v)
+        for k, v in output_stack_properties['spacing'].items()})
+    print("- origin: ", {k: float(v)
+        for k, v in output_stack_properties['origin'].items()})
+    # print(f"- chunksize: {fuse_kwargs.get('output_chunksize', None)}")
+
+    # Create the Zarr array store on disk
+    output_zarr_array = zarr.create(
+        shape=[int(i) for i in full_output_shape],
+        chunks=[int(i) for i in full_output_chunksize],
+        dtype=sims[0].data.dtype,
+        store=output_zarr_url,  # The path to the directory where the store will be created
+        overwrite=True,      # Allows overwriting if the path exists
+        **zarr_array_creation_kwargs
+        if zarr_array_creation_kwargs is not None else {},
+    )
+
+    def fuse_chunk(
+            block_id,
+            osp=output_stack_properties,
+            ns_shape=ns_shape,
+            nsdims=nsdims,
+            fuse_kwargs=fuse_kwargs,
+            output_chunksize=output_chunksize,
+            output_zarr_array=output_zarr_array,
+            ):
+        """
+        Fuse a single chunk and write to zarr array.
+        """
+
+        sdims = list(osp['shape'].keys())
+
+        normalized_chunks = normalize_chunks(
+            shape=[ns_shape[dim] for dim in nsdims] + [osp['shape'][dim] for dim in sdims],
+            chunks=(1,) * len(nsdims) +tuple(output_chunksize[dim] for dim in sdims))
+
+        ns_coord = {dim: block_id[idim] for idim, dim in enumerate(nsdims)}
+
+        spatial_chunk_ind = block_id[len(nsdims):]
+
+        chunk_offset = {sdims[idim]: int(np.sum(normalized_chunks[len(nsdims) + idim][:b])) if b > 0 else 0
+                        for idim, b in enumerate(spatial_chunk_ind)}
+        chunk_offset_phys = {dim: chunk_offset[dim] * osp['spacing'][dim] + osp['origin'][dim]
+                            for idim, dim in enumerate(sdims)}
+        chunk_shape = {sdims[idim]: normalized_chunks[len(nsdims) + idim][b]
+                    for idim, b in enumerate(spatial_chunk_ind)}
+
+        fused = fuse(
+            sims=fuse_kwargs.get("sims"),
+            **{k: v for k, v in fuse_kwargs.items() if k != "sims"},
+            output_origin={dim: chunk_offset_phys[dim] for dim in sdims},
+            output_shape={dim: chunk_shape[dim] for dim in sdims},
+            output_spacing={dim: osp['spacing'][dim] for dim in sdims},
+            ).data
+
+        fused = fused[tuple(slice(ns_coord[dim], ns_coord[dim] + 1) for dim in nsdims)]
+
+        with dask_config.set(scheduler='single-threaded'):
+            da.to_zarr(
+                fused, output_zarr_array,
+                region=tuple(
+                    [slice(ns_coord[dim], ns_coord[dim] + 1) for dim in nsdims] +
+                    [slice(chunk_offset[dim], chunk_offset[dim] + chunk_shape[dim]) for dim in sdims]),
+            )
+
+        return
+    
+    nblocks = [len(nc) for nc in normalized_chunks]
+    
+    return {
+        "func": fuse_chunk,
+        "nblocks": nblocks,
+        "output_stack_properties": output_stack_properties,
+    }
+
+
+def fuse_to_zarr(*args, **kwargs):
+    """
+    Deprecated: use fuse(..., output_zarr_url=...) instead.
+    """
+    warnings.warn(
+        "fuse_to_zarr() is deprecated. Use fuse(..., output_zarr_url=<path>) instead.",
+        DeprecationWarning,
+        stacklevel=2,
+    )
+    raise RuntimeError(
+        "fuse_to_zarr() is deprecated. Please call fuse(..., output_zarr_url=<path>) instead."
+    )
+
+
+def fuse_to_multiscale_ome_zarr(*args, **kwargs):
+    """
+    Deprecated: use fuse(..., output_zarr_url=..., zarr_options={'ome_zarr': True}) instead.
+    """
+    warnings.warn(
+        "fuse_to_multiscale_ome_zarr() is deprecated. Use fuse(..., output_zarr_url=<path>, zarr_options={'ome_zarr': True}) instead.",
+        DeprecationWarning,
+        stacklevel=2,
+    )
+    raise RuntimeError(
+        "fuse_to_multiscale_ome_zarr() is deprecated. Please call fuse(..., output_zarr_url=<path>, zarr_options={'ome_zarr': True}) instead."
+    )
